// PIC17C766 programmer
//
// Used to program device in circuit by serial programming via parallel port
// note that this must be a bidirectional parallel port
//
// INTERFACE TO HARDWARE:
//
// base address
//		Bit		Function 	Notes
//		0			MCLR			1 switches on 13V MCLR, 0 switches it off
//		1			BUFEN			1 enables data buffers writing to 766, 0 for read
//		2			DATA		  	data bit to 766
//		3			TEST			1 switches on 13V TEST, 0 switches it off
//
// base address + 1
//		Bit		Function		Notes
//		6			READ DATA	Carries copy of data pin on 766
//
// base address + 2
//		Bit		Function		Notes
//		0			CLK			1 sets DCLK on 766 LOW, 0 sets it HIGH

#include <stdio.h>
#include <string.h>
#include <conio.h>
#include <stdlib.h>
#include <dos.h>
#include <sys\timeb.h>

// these are the PIC17Cxxx commands

#define RESET_POINTER		 	0
#define LOAD_DATA 				2
#define READ_DATA 				4
#define INCREMENT_ADDRESS 		6
#define BEGIN_PROGRAMMING 		8
#define LOAD_ADDRESS			0x0a
#define READ_ADDRESS			0x0c
#define END_PROGRAMMING 	0x0e

// miscellaneous defines

typedef int BOOLEAN;

#define HIGH 1
#define LOW 0
#define TRUE 1
#define FALSE 0
#define IGNORE_BLANKS 0
#define INCLUDE_BLANKS 1

#define USER_EPROM_SIZE 0x4000
#define CONFIG_EPROM_SIZE 0x10
#define DEVICE_ID_ADDRESS 0xfdff
#define CONFIG_ADDRESS	0xfe00
#define CAL_LOOP_COUNT 50000L

// when user selects ports 1, 2 or 3 they get a port address from the following list
// note this list overwritten using values from PC BIOS RAM
#define BASE_PORT1 0x378
#define BASE_PORT2 0x278
#define BASE_PORT3 0x2bc

int baseports[] = {
	BASE_PORT1,
   BASE_PORT2,
   BASE_PORT3
};

// function definitions

int check_cfgw( void );
int load_file( FILE *fp );
int write_file( FILE *fp, unsigned short *data, unsigned short offset, unsigned short size, int blanks );
char tohex( int value );
void set_lines_low( void );
void v_prog( int );
int hexconv( int );
int program( unsigned short *data, int datalength, unsigned short *config_data );
unsigned short read_cycle();
BOOLEAN load_data( unsigned short data );
BOOLEAN program_cycle( int send_end );
BOOLEAN program_cycle_ignore_errors( int send_end );
BOOLEAN reset_address_pointer(void );
int read_device( unsigned short *u_data, unsigned short *c_data );
BOOLEAN increment_address( void );
void set_lines_low( void );
unsigned short read_cycle( void );
BOOLEAN send_command( unsigned short command );
void usec_delay( int count );
BOOLEAN send_data( unsigned short data );
BOOLEAN load_address( unsigned );
unsigned short read_data( void );
int interpret_config_data( void );
//int load_config_command( void );
#ifndef DEBUG
void getports( int * );
#endif
void null_call( void );

// global data values shared between the functions

// these are the in memory images of the user data & config data
unsigned short user_data[USER_EPROM_SIZE];
unsigned short config_data[CONFIG_EPROM_SIZE];
unsigned short device_id;

// used to hold last word read & address from chip for debug / error reporting
unsigned short data_read;
int address;
// user asks for config data to be written, this is set TRUE
int do_config = 0;
// user asks for automatic calibration option
int auto_config = 0;
// command line variable - delay loop counter for 100us program loop
int prog_loop_count = 1000;
// address of parallel port
int lpt_base = BASE_PORT1;

int datalength = 0;

char *program_errors[] = {
	"No error ?!",
   "No data in file to program",
   "Reprogram count loop exceeded ( > 25 goes )",
   "Program data verification failed",
   "Programming config word failed",
   "User abort",
   "Command or data transmission error"
};

char *file_errors[] = {
	"No error ?!!",
   "Unexpected file read error",
   "Hex data files should start with a ':'",
   "A line is too short for valid data",
   "End of data marker record is wrong length",
   "A data record is the wrong length in the file",
   "A data record checksum is wrong",
   "The end of file marker checksum is wrong",
   "Unrecognised record type present in file",
   "Address greater than size of EPROM !",
   "Non hex charcter found in input file",
   "File write error"
};

#if !defined( BUILD_READER ) && !defined( BUILD_PROGRAMMER ) && !defined( BUILD_CALIBRATOR )
#error You have not told me what to build !!!!
#endif

#if defined( BUILD_READER ) && defined( BUILD_PROGRAMMER )
#error You cannot build more than one target ( reader and programmer ! )
#endif
#if defined( BUILD_READER ) && defined( BUILD_CALIBRATOR )
#error You cannot build more than one target ( reader and calibrator ! )
#endif
#if defined( BUILD_PROGRAMMER ) && defined( BUILD_CALIBRATOR )
#error You cannot build more than one target ( programmer and calibrator ! )
#endif


main(int argc, char *argv[] )
{
	int i;
#if defined( BUILD_READER ) || defined( BUILD_PROGRAMMER )
   FILE *fp;
#endif
#ifdef BUILD_CALIBRATOR
	struct timeb timebuf, timebuf2;
   long usecs;
   long l;
#endif
#ifdef BUILD_READER
   unsigned short cksum;
   unsigned short cfgw;
   int write_all_to_file = 0;
#endif

	printf("Program version 1.1\n");
	// reset memory map images
	for( i=0; i < USER_EPROM_SIZE; i++ )
   	user_data[i] = 0xffff;
	for( i=0; i < CONFIG_EPROM_SIZE; i++ )
   	config_data[i] = 0xffff;

   // interpret command line arguments
#if defined BUILD_CALIBRATOR
	if( argc < 3 || argc > 4 )
   {
   	printf("Error - usage is 'cal-766 lpt_number delay [AUTO]'\n");
      return 1;
   }
	if( argc == 4 )
   {
   	if( strcmp( strupr(argv[3]), "AUTO" ) == 0 )
   		auto_config = 1;
	   else
   	{
   		printf("Argument 3 must be either the string 'AUTO' or must be absent\n");
	      return 2;
   	}
   }
	prog_loop_count = atoi( argv[2] );
   i = atoi( argv[1] );
#elif defined BUILD_PROGRAMMER
   if( argc < 4 || argc > 5 )
   {
   	printf("Error - usage is 'prog-766 filename lpt_number delay [CONFIG]'\n");
      return 1;
   }
	if( argc == 5 )
   {
   	if( strcmp( strupr(argv[4]), "CONFIG" ) == 0 )
   		do_config = 1;
	   else
   	{
   		printf("Argument 5 must be either the string 'CONFIG' or must be absent\n");
	      return 2;
   	}
   }
	prog_loop_count = atoi( argv[3] );
   i = atoi( argv[2] );
#elif defined BUILD_READER
   if( argc < 4 || argc > 5 )
   {
   	printf("Error - usage is 'read-766 filename lpt_number delay [ALL]'\n");
      return 1;
   }
	if( argc == 5 )
   {
   	if( strcmp( strupr(argv[4]), "ALL" ) == 0 )
   		write_all_to_file = 1;
	   else
   	{
   		printf("Argument 5 must be either the string 'ALL' or must be absent\n");
	      return 2;
   	}
   }
	prog_loop_count = atoi( argv[3] );
   i = atoi( argv[2] );
#endif

	// check that the parallel port is valid, and get its address
   if( i < 1 || i > 3 )
   {
   	printf("Error - LPT port must be a number from 1 to 3\n");
      return 1;
   }
#ifndef DEBUG
   getports( baseports );
#endif
   lpt_base = baseports[ i - 1 ];
   if( lpt_base < 0x100 || lpt_base >= 0x400 )
   {
   	printf("Er, Don't think you *have* a port %d\n", i );
      return 3;
   }
	else
   	printf("Using LPT%d ( base address 0x%4.4x )\n", i, lpt_base );

   // program code for each of the compile time options
#if defined BUILD_CALIBRATOR
	printf("Make damn sure 13V programming supply to chip is off!\n");
   printf("Hit space to start, any other key to abort...\n");
   i = getch();
	if( i != ' ' )
   	return 2;
   while( kbhit() )
   	getch();
	if( auto_config )
   {
   	printf("Calculating loop count. Please wait...\n");
      ftime(&timebuf);
		for( l=0L; l<CAL_LOOP_COUNT; l++ )
      {
			program_cycle_ignore_errors(0);
      }
      send_command(0);
      ftime(&timebuf2);
      usecs = (timebuf2.time - timebuf.time ) * 1000;
      usecs = usecs - timebuf.millitm + timebuf2.millitm;
      usecs *= 1000;
      printf("Each loop took %ld usec\n", usecs / CAL_LOOP_COUNT );
   }
   else
   {
   	printf("Trigger on Vprog, hit any key to stop\n");
   	set_lines_low();
		do
   	{
			v_prog( HIGH );
     		program_cycle_ignore_errors( 1 );
      	v_prog( LOW );
   	}
   	while( !kbhit() );
      while( kbhit() )
      	getch();
   }
   set_lines_low();
#elif defined BUILD_PROGRAMMER
   fp = fopen( argv[1], "r" );
   if( fp == NULL )
   {
   	perror("Cannot access data file ");
      return 2;
   }
   printf("Reading input file...\n");
   i=load_file( fp );
   if( i != 0 )
   {
   	printf("File input error : %s\n", file_errors[i]);
   	return i;
	}
   if( datalength == 0 )
   {
   	printf("File read but nothing to program - data file is empty!\n");
      return 3;
   }
   if( !interpret_config_data() )
   {
   	printf("Warning - Configuration values NOT VALID !\n");
      printf("Device will not be programmed\n");
      return 4;
   }
	if( config_data[15] != 0xffff )
   {
      while( kbhit() )
      	getch();
   	printf("WARNING - config data implies code protected!\n");
      printf("Which is a tad extreme, me thinks\n");
      printf("If you really want to do this, type 'P' in upper case\n");
      printf("Anything else will abort\n");
      if( getch() != 'P' )
      	return 4;
      while( kbhit() )
      	getch();
      printf("Last chance - To proceed, type 'r' in lower case\n");
      printf("Anything else to abort\n");
      if( getch() != 'r' )
      	return 4;
   }
   printf("File read, programming to address 0x%4.4x\n", datalength );
   i = program( user_data, datalength, config_data );
   set_lines_low();
   if( i == 5 )
   	while( kbhit() )
      	getch();
   if( i == 0 )
   	printf("Device programmed correctly\n");
   else
   {
      if( i < 0 )
      {
      	printf("Configuration data error at address 0x%4.4x\n", address );
      	printf("Data value 0x%4.4x, PIC17C766 data value 0x%4.4x\n", config_data[address], data_read );
			i = -i;
      }
      else
      {
   		printf("Program error at address 0x%4.4x\n", address );
      	printf("Data value 0x%4.4x, PIC17C766 data value 0x%4.4x\n", user_data[address], data_read );
   	}
      printf("%s\n", program_errors[i]);
   }
#elif  defined BUILD_READER
	printf("Reading device...\n");
	i = read_device( user_data, config_data );
//	i = hack_test();
// end of hack test
   set_lines_low();
   printf("Device ID = 0x%4.4x\n", device_id );
	switch( device_id >> 5 )
   {
   	case 9:
      	printf("Device is a PIC17C766,");
         break;
      case 0x0d:
      	printf("Device is a PIC17C762,");
         break;
      case 1:
      	printf("Device is a PIC17C756,");
         break;
      case 0x11:
      	printf("Device is a PIC17C756A,");
         break;
      case 0x15:
      	printf("Device is a PIC17C752,");
         break;
      default:
         printf("Device is of unknown type");
         break;
   }
   printf(" Revision code 0x%3.3x\n", device_id & 0x1f );
   if( i == 0 )
	{
   	printf("Device Read correctly, writing file...\n");
   	fp = fopen( argv[1], "w" );
   	if( fp == NULL )
   	{
   		perror("Cannot create data file ");
      	return 2;
   	}
   	i = write_file( fp, user_data, 0, USER_EPROM_SIZE, write_all_to_file );

     	if( i == 0 )
     		i = write_file( fp, config_data, CONFIG_ADDRESS, CONFIG_EPROM_SIZE, write_all_to_file );
      if( fputs( ":00000001FF\n", fp ) == NULL )
   		i = 11;
      if( i == 0 )
      	printf("File written correctly.\n");
      else
      	printf("File output error : %s\n", file_errors[i] );
      fclose( fp );
//	   if( !check_cfgw() )
//      	printf("Configuration data read from chip appears to be invalid.\n");
//      cksum = 0;
//      for( i=0; i < 0xfc0; i++ )
//      	cksum += user_data[ i ];
//      cksum += ( cfgw & 0x3fbd );
//      if( (cfgw & 0x1e30) == 0 )
//      	printf("Device is code protected\n");
//      else if( cksum == 0x2ffd )
//      	printf("Appears to be a blank, EPROM device with code protect off\n");
//      else if( cksum == 0x0e7d )
//      	printf("Appears to be a blank, OTP device with code protect off\n");
//      else
//      	printf("Whatever it contains, it ain't blank! Checksum = 0x%4.4x\n", cksum );
//      cksum = 0;
//      for( i=0xfc0; i < USER_EPROM_SIZE; i++ )
//      	cksum += user_data[ i ];
//      if( cksum == 0 )
//      	printf("Calibration data checksum indicates that it appears to be a blank\n");
//      else
//      	printf("Calibration data checksum is 0x%4.4x - some funny programming here\n", cksum);
   }
   else if( i > 0 )
   {
   	printf("Device read error at address 0x%4.4x\n", address );
   	printf("%s\n", program_errors[i]);
   }
   else
   {
   	printf("Device read error at config address 0x%4.4x\n", address );
   	printf("%s\n", program_errors[-i]);
   }

#else
	printf("I am a dummy program that did nothing\n");
#endif
   return i;
}

#if defined BUILD_PROGRAMMER
interpret_config_data( void )
{
	unsigned config_word;
   unsigned int i;

   config_word = config_data[0];
   for( i=1; i < CONFIG_EPROM_SIZE; i++ )
   	if( config_data[i] != 0xffff )
      	return 0;
	i = config_word & 0xbf00;
   if( i == 0 )
   	config_data[15] = 0;
   else if( i != 0xbf00 )
   	return 0;
   if( (config_word & 0x4000 ) == 0 )
   	config_data[14] = 0;
   if( (config_word & 0x0040 ) == 0 )
   	config_data[6] = 0;
   if( (config_word & 0x0010 ) == 0 )
   	config_data[4] = 0;
   if( (config_word & 0x0008 ) == 0 )
   	config_data[3] = 0;
   if( (config_word & 0x0004 ) == 0 )
   	config_data[2] = 0;
   if( (config_word & 0x0002 ) == 0 )
   	config_data[1] = 0;
   if( (config_word & 0x0001 ) == 0 )
   	config_data[0] = 0;
   else
   	config_data[0] = 0xffff;
	if( (i == 0) && ( (config_word & 0x0050 ) != 0 ) )
   	return 0;
   if( ( i != 0 ) && ( (config_word & 0x0050 ) == 0 ) )
   	return 0;
   return 1;
}
#endif

#if defined BUILD_PROGRAMMER
load_file( FILE *fp )
{
	char line[128];
   int i, length, cksum, type, len;
   unsigned short addr_base;
   unsigned long address;
   unsigned short sar = 0;
   unsigned short lba = 0;
	union hilow_ {
   	unsigned short word;
      unsigned char byte[2];
   } *bytes_ptr;
   unsigned char byte;
   datalength = 0;
   do
   {
		if( fgets( line, sizeof( line ), fp ) == NULL )
      	return 1;
      len = strlen( line );
      if( len < 12 )
      	return 3;
      if( *line != ':' )
      	return 2;
      for( i=1; i < len-1; i++ )
      	if( hexconv( line[i] ) == -1 )
         	return 10;
      length = (hexconv( line[1] ) << 4 ) + hexconv( line[2] );
      cksum = length;
      addr_base = 0;
      for( i=0; i < 4; i++ )
      	addr_base = ( addr_base << 4 ) + hexconv( line[3+i] );
//      if( ( (address >= USER_EPROM_SIZE) && (address < (unsigned)CONFIG_ADDRESS) )
//           ||
//           ( address > (unsigned)(CONFIG_ADDRESS + CONFIG_EPROM_SIZE) ) )
//           return 9;
      cksum += ( addr_base >> 8 ) + addr_base;
      type = (hexconv( line[7] ) << 4 ) + hexconv( line[8] );
      cksum += type;
      switch( type )
      {
			case 0:
		      if( strlen( line ) < ( 9 + length * 2 + 2 ) )
   		   	return 5;
      		for( i=0; i< length; i++ )
	      	{
					if( lba > 0 )
               	address = ((unsigned long)lba << 16 ) + addr_base + i;
               else if( sar > 0 )
               	address = ((unsigned long)sar << 4 ) + ( ( addr_base + i ) & 0xffff );
               else
               	address = addr_base + i;
               byte = (hexconv( line[9+i*2] ) << 4 ) + hexconv( line[10+i*2] );
               if( address >= CONFIG_ADDRESS )
               	bytes_ptr = (union hilow_ *)&config_data[(address/2-CONFIG_ADDRESS)];
               else
               	bytes_ptr = (union hilow_ *)&user_data[(address)/2];
   	   		bytes_ptr->byte[ (address) & 1 ] = byte;
               if(((address)<USER_EPROM_SIZE) && ( byte != 0xff ) && ( ( address )/2 >= datalength ) )
               	datalength = (address)/2;
	      	   cksum += byte;
		      }
   		   cksum += (hexconv( line[9+i*2] ) << 4 ) + hexconv( line[10+i*2] );
      		cksum &= 0xff;
		      if( cksum != 0 )
   		   	return 6;
      		break;
         case 1:
         	if( strlen( line ) != 12 )
            	return 4;
            cksum += (hexconv( line[9] ) << 4 ) + hexconv( line[10] );
            cksum &= 0xff;
            if( cksum != 0 )
            	return 7;
            return 0;
			case 2:
         	sar = 0;
		      for( i=0; i < 4; i++ )
      			sar = ( sar << 4 ) + hexconv( line[9+i] );
		      cksum += ( sar >> 8 ) + sar;
   		   cksum += (hexconv( line[13] ) << 4 ) + hexconv( line[14] );
      		cksum &= 0xff;
		      if( cksum != 0 )
   		   	return 6;
      		break;
         case 4:
         	lba = 0;
		      for( i=0; i < 4; i++ )
      			lba = ( lba << 4 ) + hexconv( line[9+i] );
		      cksum += ( lba >> 8 ) + lba;
   		   cksum += (hexconv( line[13] ) << 4 ) + hexconv( line[14] );
      		cksum &= 0xff;
		      if( cksum != 0 )
   		   	return 6;
         	break;
         default:
         	return 8;
      }
   }
   while( !feof( fp ) && !ferror( fp ) );
   return 1;
}
#endif // defined BUILD_PROGRAMMER

#if defined BUILD_READER
write_file( FILE *fp, unsigned short *data, unsigned short offset, unsigned short size, int include_blanks )
{
	char line[128];
   int i, length, cksum;
   int low, high;
   unsigned short address, address2, address3;
   unsigned char byte;
   char *cptr;
   int not_ff;

	address3 = (offset >> 15 );
   strcpy( line, ":02000004" );
   line[9] = tohex( (address3 >> 12 ) & 0x0f );
   line[10] = tohex( (address3 >> 8 ) & 0x0f );
   line[11] = tohex( (address3 >> 4 ) & 0x0f );
   line[12] = tohex( address3 & 0x0f );
   cksum = 2 + 4 + ( address3 >> 8 ) + ( address3 & 0xff );
   cksum = ~cksum + 1;
   line[13] = tohex( ( cksum >> 4 ) & 0xf );
   line[14] = tohex( cksum & 0xf );
	line[15] = '\n';
   line[16] = '\0';
   if( fputs( line, fp ) == NULL )
     	return 11;

   address = 0;
   while( address < size )
   {
   	for( not_ff = 0, i=0, low=8, high = 0; (i < 8) && address+i < size; i++ )
      	if( include_blanks || data[address+i] != 0xffff )
         {
         	not_ff = 1;
            if( low > i ) low = i;
            if( high < i ) high = i;
         }
      if( low < 2 )
      	low = 0;
//      if( high > 5 )
//      	high = 7;
      if( not_ff )
      {
      	// write it out
         line[0] = ':';
         length = (high - low  + 1 ) * 2;
         cksum = length;
         line[1] = tohex( length / 16 );
         line[2] = tohex( length % 16 );
         address2 = ((address + low) * 2) + (offset<<1);
         cksum += address2  >> 8;
         cksum += address2 & 0xff;
         line[3] = tohex( (address2 >> 12 ) & 0xf);
         line[4] = tohex( (address2 >> 8 ) & 0xf);
         line[5] = tohex( (address2 >> 4 ) & 0xf);
         line[6] = tohex( (address2 ) & 0xf);
         line[7] = '0';
         line[8] = '0';
			// now do the data itself
         cptr = &line[9];
         for( i = low; i <= high; i++ )
         {
            byte = (unsigned char)(data[ address + i ] & 0xff);
            cksum += byte;
            *cptr++ = tohex( (byte >> 4 ) & 0xf );
            *cptr++ = tohex( byte & 0xf );
            byte = (unsigned char)(data[ address + i ] >> 8);
            cksum += byte;
            *cptr++ = tohex( (byte >> 4 ) & 0xf );
            *cptr++ = tohex( byte & 0xf );
         }
         cksum = ~cksum + 1;
         *cptr++ = tohex( ( cksum >> 4 ) & 0xf );
         *cptr++ = tohex( cksum & 0xf );
			*cptr++ = '\n';
         *cptr = '\0';
         if( fputs( line, fp ) == NULL )
         	return 11;
      }
      address += 8;
   }
   return 0;
}
#endif // defined BUILD_READER

hexconv( character )
{
	int x;
	x = character - '0';
   if( x > 9  )
   	x = character - 'A' + 10;
   if( x < 0 || x > 15 )
   	return -1;
   return x;
}

char tohex( int value )
{
	if( value < 10 )
   	return '0' + value;
   else if( value < 16 )
   	return 'A' - 10 + value;
   else
   	return '?';
}


#if defined BUILD_PROGRAMMER
program( unsigned short *u_data, int datalength, unsigned short *c_data )
{
   int n;

   if( datalength < 1 || datalength > 0x3fff )
   	return 1;

   set_lines_low();
	v_prog( HIGH );
   // wait 80 cpu cycles
   usec_delay( 80 );

	printf("To abort printing, hit any key.\n");
   printf("Programming user data area...\n");

   if( reset_address_pointer() == FALSE )
   {
      	v_prog(LOW);
         return 6;
   }
   for( address = 0; address <= datalength; address++ )
   {
   	n = 0;
      if( load_data( u_data[address] ) == FALSE )
      {
      	v_prog(LOW);
         return 6;
      }
      do
      {
      	program_cycle( 0 );
         data_read = read_cycle();
         n++;
         if( n == 25 )
         {
         	v_prog( LOW );
            return 2;
         }
      } while( data_read != u_data[address] );
      n *= 3;
      while( n-- > 0 )
      	program_cycle( 1 );
      if( increment_address() == FALSE )
      {
      	v_prog(LOW);
         return 6;
      }
      if( kbhit() )
      {
      	v_prog( LOW );
         return 5;
      }
   }
   printf("Verifying user data area...\n");
   if( reset_address_pointer() == FALSE )
   {
     	v_prog(LOW);
      return 6;
   }
   for( address=0; address < datalength; address++ )
   {
   	data_read = read_cycle();
   	if( data_read != u_data[address] )
      {
      	v_prog( LOW );
         return 3;
      }
      if( increment_address() == FALSE )
      {
      	v_prog(LOW);
         return 6;
      }
   }
	if( do_config )
   {
   	printf("Programming config data...\n");
		load_address( CONFIG_ADDRESS );
	   for( address = 0; address < CONFIG_EPROM_SIZE; address++ )
   	{
      	if( c_data[address] != 0xffff )
   		{
         	if( c_data[address] != 0 )
            	printf("Internal error in config data, address 0x%4.4x, data 0x%4.4x\n", address, c_data[address] );
            else
            {
      			n = 100;
	      		if (load_data( c_data[address] ) == FALSE )
      			{
      				v_prog(LOW);
         			return -6;
	      		}
      			while( n-- > 0 )
      				program_cycle( 1 );
				}
	      }
   	   if( increment_address() == FALSE )
      	{
      		v_prog(LOW);
         	return -6;
      	}
      	if( kbhit() )
	      {
   	   	v_prog( LOW );
      	   return -5;
	      }
   	}
   }
   v_prog( LOW );
   return 0;
}
#endif // defined BUILD_PROGRAMMER

#if defined BUILD_READER
int hack_test()
{
	do
   {
   	set_lines_low();
		v_prog( HIGH );

     	// wait 80 cpu cycles
   	usec_delay( 80 );

   	reset_address_pointer();
		read_cycle();
   	increment_address();
		v_prog( LOW );
   }while (!kbhit() );
   while( kbhit() )
   	getch();
   return 5;
}

int read_device( unsigned short *u_data, unsigned short *c_data )
{
   set_lines_low();
	v_prog( HIGH );

   // wait 80 cpu cycles
   usec_delay( 80 );

   if( reset_address_pointer() == FALSE )
   {
   	v_prog( LOW );
     	return 6;
   }
   for( address=0; address < USER_EPROM_SIZE; address++ )
   {
   	data_read = read_cycle();
   	u_data[address] = data_read;
      if( increment_address() == FALSE )
      {
      	v_prog( LOW );
         return 6;
      }
   }
	if( load_address( CONFIG_ADDRESS ) == FALSE )
   {
   	v_prog( LOW );
     	return -6;
   }
   for( address=0; address < CONFIG_EPROM_SIZE; address++ )
   {
   	data_read = read_cycle();
      c_data[address] = data_read;
      if( increment_address() == FALSE )
      {
      	v_prog( LOW );
         return -6;
      }
   }
   if( load_address( DEVICE_ID_ADDRESS ) == FALSE )
   {
   	v_prog( LOW );
     	return -6;
   }
   device_id = read_cycle();

   v_prog( LOW );
   return 0;
}
#endif // defined BUILD_READER

int check_cfgw( )
{
	return 0;
}

// ===========================================================================
// functions to implement programming primitives of figures 4-1 to 4-7
// of the microchip programming spec

// reset address pointer - immplements figure 4-1
// debugged
BOOLEAN reset_address_pointer()
{
	BOOLEAN result;

   result = send_command( RESET_POINTER );
   // ps6 delay * safety factor
   usec_delay( 20 * 2);
	return result;
}


// increment address function - implements figure 4-2
//debugged
BOOLEAN increment_address()
{
	BOOLEAN result;

   result = send_command( INCREMENT_ADDRESS );
   // ps6 delay * safety factor
   usec_delay( 20 * 2 );
	return result;
}


//  send address - implements figure 4-3
BOOLEAN load_address( unsigned address )
{
	BOOLEAN result;

   result = send_command( LOAD_ADDRESS );
   // ps7 * safety factor
   usec_delay( 20 * 2);
	if( result )
   	result = send_data( address);
   return result;
}


// read address function - implements figure 4-4
unsigned short read_address()
{
   send_command( READ_ADDRESS );
   // ps8 delay  * safety factor
   usec_delay( 30 * 2 );
	return read_data();
}


// implements load data - figure 4-5
BOOLEAN load_data( unsigned short data )
{
	BOOLEAN result;

   result = send_command( LOAD_DATA );
	// ps7 implemented by send_data
	if( result )
   	result = send_data( data );
   return result;
}


// read cycle - implements figure 4-6
// debugeed
unsigned short read_cycle()
{
   send_command( READ_DATA );
   // ps8/9 implemented by read_data
	return read_data();
}

// implements begin programming - figure 4-7
// debugged
BOOLEAN program_cycle( int send_end )
{
   if( !send_command( BEGIN_PROGRAMMING ))
   	return FALSE;
   usec_delay( 100 );		// need 100 us prog time
	if( send_end )
	   send_command( END_PROGRAMMING );
   return TRUE;
}

// implements begin programming - figure 4-7
// debugged
BOOLEAN program_cycle_ignore_errors( int send_end )
{
   send_command( BEGIN_PROGRAMMING );
   usec_delay( 100 );		// need 100 us prog time
	if( send_end )
	   send_command( END_PROGRAMMING );
   return TRUE;
}


// ==========================================================================
// additional functions needed by high level commands

// set all lines low ( clr off, test off, data low, data read, clk low )
// implements startup function to set I/O to known state
// debugged
void set_lines_low()
{
	int i;

   outp( lpt_base, 0 );
   i = inp( (lpt_base+2) );
   outp( (lpt_base+2), (( i & 0x1f ) | 1 ));
	return;
}

// implement function to turn on or off programming voltages
// according to figure 4-8
// debugged
void v_prog( state )
{
	unsigned char i;

   i = inp( (lpt_base+2) );
   // ensure clock low
   outp( (lpt_base+2), (( i & 0x1f ) | 1 ));
	if( state == HIGH )
   {
      // set test high, keep mclr low
		outp( (lpt_base), 8 );
      // ps1 delay and some extra
	   usec_delay( 3 );
      // set mclr high
		outp( (lpt_base), 9 );
      // ps2 delay
	   usec_delay( 20 );
	}
	else
   {
	   usec_delay( 5 );
      // mclr low, test still high
		outp( (lpt_base), 8 );
	   usec_delay( 5 );
      // test low
		outp( (lpt_base), 0 );
	}
  	return;
}

// ===========================================================================
// utilities to help the above functions

// 17c7xx rewritten
// send data from figure 4-3 and 4-5
// starts from start of ps7 delay

BOOLEAN send_data( unsigned short data )
{
	int i, x, y;

   x = inp( lpt_base + 2 );
   y = inp( lpt_base ) & 0xf9;
	// ps7 * safety factor
   usec_delay( 20 * 2);
   for( i=0; i < 16; i++ )
   {
      // extra delay
		usec_delay( 1 );
      // clock high
		outp( (lpt_base+2), ( x & 0x1e ) );
      // data out with buffer enabled
		outp( (lpt_base), ((data & 1) ? 6 : 2) | y  );
      // ps4 delay
		usec_delay( 1 );
      // sanity check - read back data
      if( ((inp( lpt_base+1 ) & ( 1 << 6 ) ) ? 1 : 0 ) != (data & 1 ))
      	return FALSE;
      // clock low
		outp( (lpt_base+2), ( x & 0x1f ) | 1 );
      // next data prep
	   data >>= 1;
      // ps5
		usec_delay( 1 );
   }
   // ps6 * safety factor
   usec_delay( 20 * 2 );
	return TRUE;
}

// 17Cxxx rewritten - debugged
// implements code to send commands for all figures
// starts with first edge of clock

BOOLEAN send_command( unsigned short command )
{
	int i,x,y;

   // save port for later use
   x = inp( (lpt_base+2) );
   y = inp( lpt_base ) & 0xf9;
   for( i=0; i < 8; i++ )
   {
		// enable buffer and set data bit
		outp( (lpt_base), ((command & 1) ? 6 : 2 ) | y);
      // extra delay
		usec_delay( 1 );
      // sanity check - read back data
      if( ((inp( lpt_base+1 ) & ( 1 << 6 ) ) ? 1 : 0 ) != (command & 1 ) )
      	return FALSE;
      // clock high
		outp( (lpt_base+2), ( x & 0x1e ) );
      // ps4
		usec_delay( 1 );
      // clock low
		outp( (lpt_base+2), ( x & 0x1f ) | 1 );
   	command >>= 1;
		// ps5
		usec_delay( 1 );
   }
   return TRUE;
}

// 17c7xx done - not debugged
// implements 16 bit data read part of figures 4-4 4-6
// from start of ps-9 timing delay

unsigned short read_data( void )
{
	int i;
   unsigned short data;
	unsigned char port;

   // PS9 delay - after clock low, wait 4/Fosc usec
   usec_delay( 1 );
	// now set data line to input
   port = inp( (lpt_base) );
   port &= 0xfd;
   outp( (lpt_base), port );
   // now wait (PS8 - ps9) ( 30 * 4/Fosc ) * safety factor
   usec_delay( 30 * 2 );
   data = 0;
   for( i = 0; i < 16; i++ )
   {
		// set clock high
	   outp( (lpt_base+2), ( port & 0x3e ) );
		// ps4 delay
	   usec_delay( 1 );
      // read data bit
	   data = ( data >> 1 ) | (( inp( lpt_base+1 ) & (1<<6) ) ? 0x8000 : 0 );
      // clock low
	   outp( (lpt_base+2), ( port & 0x3f ) | 1 );
      // ps5 delay
	   usec_delay( 1 );
   }
   // ps6 delay * safety factor
   usec_delay( 20 * 2 );
   port = inp( (lpt_base) );
   port |= 2;
   outp( (lpt_base), port );
	return data;
}

// 17Cxxx OK - patch out IO routines

#ifdef DEBUG
outp( int address, int value )
{
	return 1;
}
inp( int address )
{
	return 1;
}
#endif


// 17Cxxx OK - get IO port addresses
#ifndef DEBUG
void getports( int *ports )
{
	int i;
   int far *ptr;

   ptr = MK_FP( 0x40, 8 );
   for( i=0; i < 3; i++ )
   	ports[i] = ptr[i];
}
#endif

void usec_delay( int microseconds )
{
	unsigned long delay;

   delay = microseconds;
   delay *= prog_loop_count;
	while( delay-- )
   	null_call();
}


void null_call()
{
}


